import 'dart:developer';
import 'dart:io';

import 'package:health/health.dart';

import 'package:vivanta_connect_flutter/enums/apple_health_parameter_type.dart';
import 'package:vivanta_connect_flutter/enums/google_health_parameter_type.dart';
import 'package:vivanta_connect_flutter/helpers/helpers.dart';
import 'package:vivanta_connect_flutter/models/apple_health_element.dart';
import 'package:vivanta_connect_flutter/models/google_health_connect_element.dart';
import 'package:vivanta_connect_flutter/services/apple_health.dart';
import 'package:vivanta_connect_flutter/services/database.dart';
import 'package:vivanta_connect_flutter/services/google_health_connect.dart';
import 'package:vivanta_connect_flutter/services/http.dart';
import 'package:vivanta_connect_flutter/services/preferences.dart';

class VivantaSync {
  final String apiKey;
  final String customerId;
  final String externalUserId;
  late final Database database;
  late final Http http;
  late final Preferences preferences;

  VivantaSync({
    required this.apiKey,
    required this.customerId,
    required this.externalUserId,
  }) {
    http = Http();
    database = Database();
    preferences = Preferences();
  }

  Future<void> executeAll() async {
    await database.init();

    if (Platform.isIOS) {
      await executeAppleHealth(database, preferences, http);
    }
    if (Platform.isAndroid) {
      await executeGoogleHealthConnect(database, preferences, http);
    }

    return;
  }

  Future<void> executeAppleHealth(
    Database database,
    Preferences preferences,
    Http http,
  ) async {
    if (await preferences.getAppleHealthAccessGranted()) {
      await _getAppleHealthData(preferences, http);
      await _syncAppleHealthData(database);
      await _deleteOldAppleHealthElements(database);
    }
  }

  Future<void> executeGoogleHealthConnect(
    Database database,
    Preferences preferences,
    Http http,
  ) async {
    if (await preferences.getGoogleHealthConnectAccessGranted()) {
      await _getGoogleHealthConnectData(preferences, http);
      await _syncGoogleHealthConnectData(database);
      await _deleteOldGoogleHealthConnectElements(database);
    }
  }

  Future<void> _getAppleHealthData(Preferences preferences, Http http) async {
    final appleHealth = AppleHealth();

    final client = await http.me(apiKey, customerId);
    final healthPermissions = getAppleHealthPermissionFromClient(client);

    final healthDataTypes =
        healthPermissions.map((e) => e.healthDataType).toList();

    final lastSync = await preferences.getAppleHealthSyncDate();
    final todaySync = DateTime.now();
    final values = await appleHealth.getDataPoint(
      healthDataTypes,
      lastSync,
      todaySync,
    );
    for (final value in values) {
      final element = AppleHealthElement(
        id: value.sourceId,
        value: value.toJson().toString(),
        date: DateTime.now().toIso8601String(),
        synced: false,
      );
      await database.saveAppleHealthElement(element);
    }
    await preferences.setAppleHealthSyncDate(todaySync);
  }

  Future<void> _syncAppleHealthData(Database database) async {
    const elementsPerRequest = 100;
    final List<String> values = [];
    final appleHealthElements = database.getUnsyncedAppleHealthElements();

    for (int i = 0;
        i < appleHealthElements.length && i < elementsPerRequest;
        i++) {
      values.add(appleHealthElements[i].value!);
      appleHealthElements[i].synced = true;
      appleHealthElements[i].syncedDate = DateTime.now().toIso8601String();
    }

    final result = await _uploadAppleHealthArrayData(http, values);

    if (result) {
      for (final element in appleHealthElements) {
        final wasSync = element.synced ?? false;
        if (wasSync) {
          await database.updateAppleHealthElement(element);
        }
      }
    }

    if (appleHealthElements.length > elementsPerRequest && result) {
      await _syncAppleHealthData(database);
    }
  }

  Future<bool> _uploadAppleHealthArrayData(Http http, List<String> value) {
    return http.uploadAppleHealthArrayData(
      apiKey,
      customerId,
      externalUserId,
      value,
    );
  }

  Future<void> _deleteOldAppleHealthElements(Database database) async {
    final result = database.getOldSyncedAppleHealthElements();
    for (final element in result) {
      database.deleteAppleHealthElement(element);
    }
  }

  Future<void> _getGoogleHealthConnectData(
      Preferences preferences, Http http) async {
    final healthConnect = GoogleHealthConnect();

    if (!await healthConnect.isInstalled()) {
      log("Google Health Connect is not installed");
      return;
    }

    final client = await http.me(apiKey, customerId);
    final healthPermissions = getGoogleHealthPermissionFromClient(client);

    final healthDataTypes =
        healthPermissions.map((e) => e.healthDataType).toList();

    final lastSync = await preferences.getGoogleHealthConnectSyncDate();
    final todaySync = DateTime.now();
    final values = await healthConnect.getDataPoint(
      healthDataTypes,
      lastSync,
      todaySync,
    );
    for (final value in values) {
      final element = GoogleHealthConnectElement(
        id: value.sourceId,
        value: value.toJson().toString(),
        date: DateTime.now().toIso8601String(),
        synced: false,
      );
      await database.saveGoogleHealthConnectElement(element);
    }
    await preferences.setGoogleHealthConnectSyncDate(todaySync);
  }

  Future<void> _syncGoogleHealthConnectData(Database database) async {
    const elementsPerRequest = 100;
    final List<String> values = [];
    final googleHealthConnectElements =
        database.getUnsyncedGoogleHealthConnectElements();

    for (int i = 0;
        i < googleHealthConnectElements.length && i < elementsPerRequest;
        i++) {
      values.add(googleHealthConnectElements[i].value!);
      googleHealthConnectElements[i].synced = true;
      googleHealthConnectElements[i].syncedDate =
          DateTime.now().toIso8601String();
    }

    final result = await _uploadGoogleHealthConnectArrayData(http, values);

    if (result) {
      for (final element in googleHealthConnectElements) {
        final wasSync = element.synced ?? false;
        if (wasSync) {
          await database.updateGoogleHealthConnectElement(element);
        }
      }
    }

    if (googleHealthConnectElements.length > elementsPerRequest && result) {
      await _syncGoogleHealthConnectData(database);
    }
  }

  Future<bool> _uploadGoogleHealthConnectArrayData(
      Http http, List<String> value) {
    return http.uploadGoogleHealthConnectArrayData(
      apiKey,
      customerId,
      externalUserId,
      value,
    );
  }

  Future<void> _deleteOldGoogleHealthConnectElements(Database database) async {
    final result = database.getOldSyncedGoogleHealthConnectElements();
    for (final element in result) {
      database.deleteGoogleHealthConnectElement(element);
    }
  }

  Future<List<AppleHealthParameterType>>
      getAppleHealthWritePermissions() async {
    final client = await http.me(apiKey, customerId);
    final healthPermissions = getAppleHealthPermissionFromClient(client);

    List<AppleHealthParameterType> parameterTypes = [];

    for (final permission in healthPermissions) {
      if (permission.healthDataAccess == HealthDataAccess.READ_WRITE ||
          permission.healthDataAccess == HealthDataAccess.WRITE) {
        final parameterType =
            AppleHealth.getParameterFromDataType(permission.healthDataType);
        if (!parameterTypes.any((element) => element == parameterType)) {
          parameterTypes.add(parameterType);
        }
      }
    }

    return parameterTypes;
  }

  Future<List<GoogleHealthParameterType>>
      getGoogleHealthConnectWritePermissions() async {
    final client = await http.me(apiKey, customerId);
    final healthPermissions = getGoogleHealthPermissionFromClient(client);

    List<GoogleHealthParameterType> parameterTypes = [];

    for (final permission in healthPermissions) {
      if (permission.healthDataAccess == HealthDataAccess.READ_WRITE ||
          permission.healthDataAccess == HealthDataAccess.WRITE) {
        final parameterType = GoogleHealthConnect.getParameterFromDataType(
            permission.healthDataType);
        if (!parameterTypes.any((element) => element == parameterType)) {
          parameterTypes.add(parameterType);
        }
      }
    }

    return parameterTypes;
  }
}
